/*******************************************************************************
 * Copyright 2012 University of Southern California
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * 	http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * This code was developed by the Information Integration Group at 
 * the Information Sciences Institute of the University of Southern California.  
 * For more information, publications, and related projects, 
 * please see: http://www.isi.edu/integration
 ******************************************************************************/

package mf.classify;

import java.util.ArrayList;

import mf.common.Constants;
import mf.ftrs.WaterFilling;

public class KNN {

	private final static int N = 9 ;

	/**
	 * @param wf The WaterFilling object for an image
	 * @param featuresList List of feature vectors of 24 dimensions
	 * @return A list of N closest neighbors
	 * @throws Exception
	 * This method compares the given features for an image (contained in the WaterFilling object) with 
	 * all the features in the given featuresList.
	 * It then returns N closest neighbors based on a metric. The metric we use is the L1-distance between the vectors.
	 */
	public static ArrayList<Integer> nearestNeighbors(WaterFilling wf, ArrayList<int[]> featuresList) throws Exception {
		ArrayList<Integer> leastDistances ;
		ArrayList<Integer> closestNeighbors ;
		
		leastDistances = new ArrayList<Integer>() ;
		closestNeighbors = new ArrayList<Integer>() ;
		// Initialize the ArrayLists.
		// Add max int value to the distance list and -1 to the neighbors list
		for(int i=0;i<N;i++) {
			leastDistances.add(Integer.MAX_VALUE) ;
			closestNeighbors.add(-1) ;
		}
		// compare with all feature vectors 
		for(int nbrIndx=0;nbrIndx<featuresList.size();nbrIndx++) {
			int distance;
			int[] featureVector;
			featureVector = featuresList.get(nbrIndx) ;
			distance = 0 ;
			for(int i=0;i<8;i++) { 
				distance+= Math.abs(wf.FTH[i][0] - featureVector[i]) ;
			}
			for(int i=0;i<8;i++) {
				distance+= Math.abs(wf.FCH[i][0] - featureVector[i+8]) ;
			}
			for(int i=0;i<8;i++) {
				distance+= Math.abs(wf.WAH[i]    - featureVector[i+16]) ;
			}
			if(distance < leastDistances.get(N-1)) {
				int index = N-2 ;
				while(index>=0 && distance < leastDistances.get(index)) {
					index-- ;
				}
				leastDistances.remove(N-1) ;
				closestNeighbors.remove(N-1) ;
				leastDistances.add(index+1, distance) ;
				closestNeighbors.add(index+1, nbrIndx) ;
			}
		}
		return closestNeighbors ;
	}
	
	
	
	/**
	 * @param wf The WaterFilling object containing the image features
	 * @param namesList The list of names in the index
	 * @param featuresList The list of their features in the index
	 * @return A constant indicating if the predicted type for the given image is map or non-map
	 * @throws Exception
	 */
	public static int predictType(WaterFilling wf, ArrayList<String> namesList, ArrayList<int[]> featuresList) throws Exception {
		ArrayList<Integer> closestNeighbors ;
		int numMatchingMaps, numMatchingNonmaps ; 
		
		numMatchingMaps = numMatchingNonmaps = 0 ; 
		closestNeighbors = nearestNeighbors(wf, featuresList) ;
		for (int nbrIndex : closestNeighbors) {
			if(State.label(namesList.get(nbrIndex)) == Constants.MAP) {
				numMatchingMaps++ ;
			}
			else {
				numMatchingNonmaps++ ;
			}
		}
		if(numMatchingMaps > numMatchingNonmaps) { 
			return Constants.MAP ;
		}
		else { 
			return Constants.NONMAP ;
		}
	}
	
}


