/*******************************************************************************
 * Copyright 2012 University of Southern California
 * 
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 * 
 * 	http://www.apache.org/licenses/LICENSE-2.0
 * 
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 * 
 * This code was developed by the Information Integration Group at 
 * the Information Sciences Institute of the University of Southern California.  
 * For more information, publications, and related projects, 
 * please see: http://www.isi.edu/integration
 ******************************************************************************/

package mf.ftrs;

import java.awt.image.BufferedImage;
import java.awt.image.PixelGrabber;

import mf.common.Constants;

public class WaterFilling {

	int width, height, numSegments ;
	int[] MFTFC, MFCFT, coloredPixels ;
	int[][] imagePixels, segmentFeatures ;
	
	public int[][] FTH, FCH;
	public int[] WAH;
	

	/**
	 * @param bimg The BufferedImage object that contains the image read from disk.
	 * @throws Exception
	 * This constructor is the only public method of this class. 
	 * It analyzes the given image, extracts features for each segment in it,
	 * creates histograms of distributions of the 3 segment specific features -
	 * Filling Time (FT), Fork Count (FC) and Water Amount (WA).
	 */
	public WaterFilling(BufferedImage bimg) throws Exception {
		int numPixels, numBlackPixels;
		PixelGrabber pg ;			
		int[] wfront ;
		int[] tmpWfront ;
		int nWfront ;
		int numTrees ;
		
		MFTFC = new int[2];
		MFCFT = new int[2];
		FTH=new int[Constants.numBins][2];
		FCH=new int[Constants.numBins][2];
		WAH=new int[Constants.numBins];
		
		width = bimg.getWidth();
		height = bimg.getHeight() ;
		numPixels = width * height ;
		coloredPixels = new int[numPixels] ;
		imagePixels = new int[numPixels][2] ;
		pg = new PixelGrabber(bimg, 0, 0, width, height, coloredPixels, 0, width) ;
		pg.grabPixels() ;
		numBlackPixels = 0 ;
		// Coverting a grey scale image into black and white representation
		// imagePixels represents black pixels with 0 and white pixels with 1
		// The second index in the imagePixels stores whether this black pixel has water filled in it.
		for(int i=0;i<numPixels;i++) {
			int pixelColor = coloredPixels[i] & 0xffffff ;
			if(pixelColor > 0) 
				imagePixels[i][0] = 1 ;
			else {
				imagePixels[i][0] = 0 ;
				imagePixels[i][1] = 0 ;
				numBlackPixels++ ;
			}
		}
		segmentFeatures = new int[numBlackPixels][3] ;
		numSegments = 0 ;
		wfront = new int[numBlackPixels] ;
		tmpWfront = new int[numBlackPixels] ;
		// Run thru all pixels
		for(int i=0;i<numPixels;i++) {
			int ft,fc,wa ;
			int[] nbrs ;
			// if pixel is white or it is black but already filled with water, 
			// continue on to next pixel
			if(imagePixels[i][0] == 1 || imagePixels[i][1] == 1) {
				continue ;
			}
			ft = 0 ; 
			fc = 0 ; 
			wa = 1 ;
			nWfront = 1 ;
			// record that this pixel is now filled
			imagePixels[i][1] = 1 ;
			wfront[0] = i ;
			while(nWfront > 0) {
				int tmpNWfront ;
				tmpNWfront = 0 ;
				for(int f=0;f<nWfront;f++) {
					nbrs = possibleNextPxl(wfront[f]) ;
					for(int n=1;n<=nbrs[0];n++) {
						tmpWfront[tmpNWfront] = nbrs[n] ;
						tmpNWfront++ ;
						imagePixels[nbrs[n]][1] = 1 ;
						wa++ ;
					}
					if(nbrs[0] > 0) {
						fc+=(nbrs[0] -1) ;
					}
						
					if(nbrs[0] > 1) {
						makeDisplayArray(wfront[f], 1) ;
					}
					else { 
						makeDisplayArray(wfront[f], 0) ;
					}
				}
				ft++ ;
				for(int count=0;count<tmpNWfront;count++) { 
					wfront[count] = tmpWfront[count] ;
				}
				nWfront = tmpNWfront ;
			}
			segmentFeatures[numSegments][0] = ft ; 
			segmentFeatures[numSegments][1] = fc ;
			segmentFeatures[numSegments][2] = wa ;
			numSegments++ ;
		}
		buildHistogramAndFindMaxFeatures();
		numTrees = 0 ;
		MFTFC[0] = 0 ;
		MFTFC[1] = 0 ;
		MFCFT[0] = 0 ;
		MFCFT[1] = 0 ;
		for(int i=0;i<Constants.numBins;i++) { 
			numTrees+=FTH[i][0] ;
		}
		double ratio = 1000.00 / ((double) numTrees) ;
		for(int i=0;i<Constants.numBins;i++) {
			FTH[i][0] = (int) Math.round(FTH[i][0] * ratio) ;
			FTH[i][1] = 0 ;
			FCH[i][0] = (int) Math.round(FCH[i][0] * ratio) ;
			FCH[i][1] = 0 ;
			WAH[i] = (int) Math.round(WAH[i] * ratio) ;
		}
	}

	
	
	/**
	 * @param row The row index of the pixel in the image
	 * @param col The column index of the pixel in the image
	 * @param position Position relative to the pixel (explained below)
	 * @return The index of the neighboring pixel.
	 * @throws Exception
	 * The argument - position, indicated the position of the neighoring pixel. 
	 * Its value is a number between 1 and 9, where the pixel is assumed to be located at box number 5.
	 *    1 |  2 | 3
	 *    -----------
	 *    4 |  5 | 6
	 *    -----------
	 *    7 |  8 | 9 
	 * So, if the index of the pixel above the pixel given by (row, col) is reqd, then the position argument should have the value - 2.
	 * Since the image has been flattened into a 1-dimensional array, this method returns the index of the required pixel in the 1-dim array.  
	 */
	private int index(int row, int col,int position) throws Exception {
		int i;

		i = 0 ;
		switch (position) {
		case 1: i = (row-1)*width+col-1 ; break ;
		case 2: i = (row-1)*width+col   ; break ;
		case 3: i = (row-1)*width+col+1 ; break ;
		case 4: i = (row  )*width+col-1 ; break ;
		case 6: i = (row  )*width+col+1 ; break ;
		case 7: i = (row+1)*width+col-1 ; break ;
		case 8: i = (row+1)*width+col   ; break ;
		case 9: i = (row+1)*width+col+1 ; break ;
		default: throw new Exception("index() function in class WaterFilling got invalid position : " + position) ;
		}
		return i ;
	} 

	
	/**
	 * @param pixelIndex The index of the pixel in the 1-dim array
	 * @param unfilledOnly Flag that indicates if only unfilled pixels should be returned
	 * @return An array, the first element of which indicates how many neighbors exist in the rest of the array.
	 * @throws Exception
	 * This method returns which among the 8 neighbors of the given pixel are black.
	 */
	private int[] get8Nbrs(int pixelIndex, boolean unfilledOnly) throws Exception {
		int row;
		int col ;
		int[] nbr ;
		boolean filled ;
		int nbrPixelIndex ;
		
		nbr = new int[9] ;
		row = pixelIndex/width;
		col = pixelIndex%width;
		filled = !unfilledOnly ;
		nbr[0] = 0 ;
		nbrPixelIndex = index(row,col,2) ;
		if(row!=0 && imagePixels[nbrPixelIndex][0] == 0 && (filled || imagePixels[nbrPixelIndex][1] == 0)) {
			nbr[0]++ ;
			nbr[nbr[0]] = nbrPixelIndex ;
		}
		nbrPixelIndex = index(row,col,4) ;
		if(col!=0 && imagePixels[nbrPixelIndex][0] == 0 && (filled || imagePixels[nbrPixelIndex][1] == 0)) {
			nbr[0]++ ;
			nbr[nbr[0]] = nbrPixelIndex ;
		}
		nbrPixelIndex = index(row,col,6) ;
		if(col!=(width-1) && imagePixels[nbrPixelIndex][0] == 0 && (filled || imagePixels[nbrPixelIndex][1] == 0)) {
			nbr[0]++ ;
			nbr[nbr[0]] = nbrPixelIndex ;
		}
		nbrPixelIndex = index(row,col,8) ; 
		if(row!=(height-1) && imagePixels[nbrPixelIndex][0] == 0 && (filled || imagePixels[nbrPixelIndex][1] == 0)) {
			nbr[0]++ ;
			nbr[nbr[0]] = nbrPixelIndex ;
		}
		nbrPixelIndex = index(row,col,1) ;
		if(row!=0 && col!=0 && imagePixels[nbrPixelIndex][0] == 0 && (filled || imagePixels[nbrPixelIndex][1] == 0)) {
			nbr[0]++ ;
			nbr[nbr[0]] = nbrPixelIndex ;
		}
		nbrPixelIndex = index(row,col,3) ;
		if(row!=0 && col!=(width-1) && imagePixels[nbrPixelIndex][0] == 0 && (filled || imagePixels[nbrPixelIndex][1] == 0)) {
			nbr[0]++ ;
			nbr[nbr[0]] = nbrPixelIndex ;
		}
		nbrPixelIndex = index(row,col,7) ;
		if(row!=(height-1) && col!=0 && imagePixels[nbrPixelIndex][0] == 0 && (filled || imagePixels[nbrPixelIndex][1] == 0)) {
			nbr[0]++ ;
			nbr[nbr[0]] = nbrPixelIndex ;
		}
		nbrPixelIndex = index(row,col,9) ;
		if(row!=(height-1) && col!=(width-1) && imagePixels[nbrPixelIndex][0] == 0 && (filled || imagePixels[nbrPixelIndex][1] == 0)) {
			nbr[0]++ ;
			nbr[nbr[0]] = nbrPixelIndex ;
		}
		return nbr;
	} 

	
	/**
	 * @param pixelIndex The index of the pixel in the 1-dim array
	 * @param unfilledOnly Flag that indicates if only unfilled pixels should be returned
	 * @return An array, the first element of which indicates how many neighbors exist in the rest of the array.
	 * @throws Exception
	 * This method returns which among the 4 (north, south, east, west only) neighbors of the given pixel are black.
	 */
	private int[] get4Nbrs(int pixelIndex, boolean unfilledOnly) throws Exception {
		int row ;
		int col ;
		int[] nbr ;
		boolean filled ;
		int nbrPixelIndex ;
		
		row = pixelIndex/width ;
		col = pixelIndex % width ;
		nbr = new int[9] ;
		filled = !unfilledOnly ;
		nbr[0] = 0 ;
		nbrPixelIndex = index(row,col,2) ;
		if(row!=0 && imagePixels[nbrPixelIndex][0] == 0 && (filled || imagePixels[nbrPixelIndex][1] == 0)) {
			nbr[0]++ ;
			nbr[nbr[0]] = nbrPixelIndex ;
		}
		nbrPixelIndex = index(row,col,4) ;
		if(col!=0 && imagePixels[nbrPixelIndex][0] == 0 && (filled || imagePixels[nbrPixelIndex][1] == 0)) {
			nbr[0]++ ;
			nbr[nbr[0]] = nbrPixelIndex ;
		}
		nbrPixelIndex = index(row,col,6) ;
		if(col!=(width-1) && imagePixels[nbrPixelIndex][0] == 0 && (filled || imagePixels[nbrPixelIndex][1] == 0)) {
			nbr[0]++ ;
			nbr[nbr[0]] = nbrPixelIndex ;
		}
		nbrPixelIndex = index(row,col,8) ; 
		if(row!=(height-1) && imagePixels[nbrPixelIndex][0] == 0 && (filled || imagePixels[nbrPixelIndex][1] == 0)) {
			nbr[0]++ ;
			nbr[nbr[0]] = nbrPixelIndex ;
		}
		return nbr ;
	} 
	

	
	/**
	 * @param pixelIndex The index of the pixel whose nbrs are needed
	 * @param unfilledOnly Should only unfilled neighbors be returned?
	 * @return Neighbors
	 * @throws Exception
	 * This method returns only those pixels that do not touch each other
	 */
	private int[] getNonTouchingNbrs(int pixelIndex, boolean unfilledOnly) throws Exception {
		int[] nbrs, nonTouchingNbrs, nbrsOfNbr ;
		
		nonTouchingNbrs = new int[9] ;
		nbrsOfNbr = new int[9] ;
		nbrs = get8Nbrs(pixelIndex, unfilledOnly) ;
		for(int k=0;k<=nbrs[0];k++) {
			nonTouchingNbrs[k] = nbrs[k] ; 
		}
		for(int i=1; i<nbrs[0]; i++) {
			nbrsOfNbr = get4Nbrs(nbrs[i],unfilledOnly) ;
			nonTouchingNbrs[0] = i;
			for(int j=i+1;j<=nbrs[0];j++) {
				int n = nbrs[j] ;
				int flag = 0 ;
				for(int k=1 ; k<=nbrsOfNbr[0] ; k++) {
					if(n == nbrsOfNbr[k]) {
						flag = 1 ;
						break ;
					}
				}
				if(flag == 0) {
					nonTouchingNbrs[0]++ ;
					nonTouchingNbrs[nonTouchingNbrs[0]] = nbrs[j] ;
				}
			}
			for(int k=0;k<=nonTouchingNbrs[0];k++) {
				nbrs[k] = nonTouchingNbrs[k] ;
			}
		}
		return nbrs ;
	} 


	/**
	 * @param pixelIndex The index of the pixel whose non-touching nbrs are neeeded.
	 * @return Non-touching neighbors
	 * @throws Exception
	 * This is a convenience method to return all non-touching nbrs irrespective of whether they are filled or not.
	 */
	private int[] getAllNonTouchingNbrs(int pixelIndex) throws Exception {
		return getNonTouchingNbrs(pixelIndex, false) ;
	} 

	
	
	/**
	 * @param pixelIndex The index of the pixel for which possible next pixels to follow are being requested
	 * @return The possible next pixels
	 * @throws Exception
	 * This method returns the possible next pixels that are neighbors of the supplied pixel,
	 * and that should be filled with water.
	 */
	private int[] possibleNextPxl(int pixelIndex) throws Exception {
		int[] nbrs ;
		int[] nonfilledNbrs ;
		
		nonfilledNbrs = new int[9] ;
		nbrs = getAllNonTouchingNbrs(pixelIndex) ;
		nonfilledNbrs[0] = 0 ;
		for(int i=1;i<=nbrs[0];i++) {
			if(imagePixels[nbrs[i]][1] == 0) {
				nonfilledNbrs[0]++ ;
				nonfilledNbrs[nonfilledNbrs[0]] = nbrs[i] ;
			}
		}
		return nonfilledNbrs ;
	} 

	
	
	 
	/**
	 * This method runs through the features recorded per segment and 
	 * creates the Filling Time (FT), Fork Count (FC) and Water Amount (WA) histograms.
	 * It also records the largest FT and FC for it (MFTFC).
	 * It also records the largest FC and FT for it (MFCFT).
	 */
	private void buildHistogramAndFindMaxFeatures()	{
		int j;

		if(numSegments == 0)
			return ;

		MFTFC[0] = -1 ;
		MFCFT[0] = -1 ;
		for(int segmentIndex=0;segmentIndex<numSegments;segmentIndex++)	{
			int ft, fc, wa;
			ft = segmentFeatures[segmentIndex][0] ;
			fc = segmentFeatures[segmentIndex][1] ;
			wa = segmentFeatures[segmentIndex][2] ;
			if(ft > MFTFC[0]) {
				MFTFC[0]=ft;
				MFTFC[1]=fc;
			}
			if(fc > MFCFT[0]) {
				MFCFT[0]=fc;
				MFCFT[1]=ft;
			}
			for(j=0;j<7 && ft>Constants.FTH_BIN[j];j++) ;
			FTH[j][0]++;
			FTH[j][1]+=fc;
			for(j=0;j<7 && fc>Constants.FCH_BIN[j];j++) ;
			FCH[j][0]++;
			FCH[j][1]+=ft;
			for(j=0;j<7 && wa>Constants.WAH_BIN[j];j++) ;
			WAH[j]++;
		}
		for(int i=0;i<8;i++){
			if(FTH[i][0]!=0) {
				FTH[i][1]=FTH[i][1]/FTH[i][0];
			}
			if(FCH[i][0]!=0) {
				FCH[i][1]=FCH[i][1]/FCH[i][0];
			}
		}
	} 

	public String toString() {
		String s = ("\nMFTFC " + MFTFC[0] + " " + MFTFC[1] + "\nMFCFT " + MFCFT[0] + " " + MFCFT[1] + "\n");
		for(int j=0;j<8;j++) {
			s+= (j + "  FTH " + " "+FTH[j][0]+" "+FTH[j][1]+"       FCH "+FCH[j][0]+" "+FCH[j][1]+"       WAH "+WAH[j] + "\n");
		}
		return s ;
	}


	private void makeDisplayArray(int p, int c) {
		if(c == 1) 
			coloredPixels[p] = 0xe1194 ;
		else 
			coloredPixels[p] = 0x9b94 ;
	}


	
}